//Chair
import toxi.geom.*;
import toxi.geom.mesh.*;
import toxi.math.*;
import toxi.volume.*;
import toxi.processing.*;

int RES = 256 ;//16,32,64,128,256
float ISO =  0;// 0.5//Unterscheidung von innen - außen, also zwischen ISO und MAX_ISO
float ISOStart = ISO;//Speichert den Startwert
float MAX_ISO = 0.3;// 2 UpperBound
boolean KZeichnen = false;//Koordinatenachsen nicht zeichnen

WETriangleMesh mesh;

ToxiclibsSupport gfx;
float currZoom = 1;
boolean isWireframe;
boolean inFarbe=false;
boolean lichtV,lichtH,lichtR,lichtL,lichtO,lichtU;
boolean umZAchseDrehen = false;


void setup() {
  size(800,800, P3D);
  colorMode(HSB);
  lichtH=true; lichtV=true; lichtO=true; lichtU=false; lichtR=false; lichtL=true;
  gfx = new ToxiclibsSupport(this);
  VolumetricSpace vol = new EvaluatingVolume(new Vec3D(800,800,800), RES, MAX_ISO);
  IsoSurface surface = new HashIsoSurface(vol);
  mesh = new WETriangleMesh();
  surface.computeSurfaceMesh(mesh, ISO);
 
}

void draw() {
  background(0);
  translate(width / 2, height / 2, 0);
  rotateX(mouseY * 0.01f);
  rotateY(mouseX * 0.01f);
  if(umZAchseDrehen){
    rotateZ(1.5);
  }
  scale(currZoom);
  colorMode(RGB);//damit die Koordinatenachsen r,g,b gezeichnet werden
  if(KZeichnen) gfx.origin(500);//Koordinatenachsen
  colorMode(HSB);
  if (isWireframe) {
    noFill();
    stroke(255);
  } 
  else {
   fill(20,220,255); 
   noStroke();
   if(lichtV) directionalLight(0, 0, 255, 0, 0, -1); 
   if(lichtH) directionalLight(0, 0, 255, 0, 0,  1); 
   if(lichtR) directionalLight(0, 0, 255,-1, 0,  0); 
   if(lichtL) directionalLight(0, 0, 255, 1, 0,  0); 
   if(lichtO) directionalLight(0, 0, 255, 0,-1,  0);
   if(lichtU) directionalLight(0, 0, 255, 0, 1,  0); 
  }
  if(!inFarbe) gfx.mesh(mesh, true);//hier wird gezeichnet //einfarbig   
   else gfx.meshNormalMapped(mesh, !isWireframe);//bunt
}

void voxelizeMesh() {
  if(ISO==0) ISO = 0.00001;//Funktioniert nicht mit ISO = 0
  mesh.computeFaceNormals();
  mesh.faceOutwards();
  mesh.computeVertexNormals();
  mesh.computeFaceNormals();
  mesh.faceOutwards();
  mesh.computeVertexNormals();
  MeshVoxelizer voxelizer=new MeshVoxelizer(RES);
  voxelizer.setWallThickness(0);//0,1,2
  VolumetricSpace vol = voxelizer.voxelizeMesh(mesh);
  vol.closeSides();
  IsoSurface surface = new HashIsoSurface(vol);
  mesh = new WETriangleMesh();
  surface.computeSurfaceMesh(mesh, ISO);
  mesh.computeVertexNormals();
  ISO = ISOStart;//zurücksetzen
}

void keyPressed() {
   if (key == 'w') {
    isWireframe = !isWireframe;
  }
  if (key == 'x') {
   voxelizeMesh();
  }  
  if (key == '-') {
    currZoom -= 0.1;
  }
  if (key == '+') {
    currZoom += 0.1;
  }
  if (key == 'l') {
    new LaplacianSmooth().filter(mesh, 1);
  }
  if (key == 'k') {
    KZeichnen= !KZeichnen;
  }
  if (key == 'f') {
    inFarbe= !inFarbe;
  }
  if (key == 'r') {
    umZAchseDrehen = true;
  }
  if (key == 's') {
     saveFrame("Bild-##.png");
  }
  if(key=='O') {lichtO=!lichtO;println("Licht oben: "+lichtO);}
  if(key=='U') {lichtU=!lichtU;println("Licht unten: "+lichtU);}
  if(key=='L') {lichtL=!lichtL;println("Licht links: "+lichtL);}
  if(key=='R') {lichtR=!lichtR;println("Licht rechts: "+lichtR);}
  if(key=='V') {lichtV=!lichtV;println("Licht vorn: "+lichtV);}
  if(key=='H') {lichtH=!lichtH;println("Licht hinten "+lichtH);}
  if (key == 'm') {
    beginShape(TRIANGLES);
      gfx.mesh( mesh );
    endShape();
    mesh.saveAsSTL( sketchPath( "cube.stl" ));
  }
}

class EvaluatingVolume extends VolumetricSpace {
  private int X,Y,Z;
  private float upperBound;
  
  public EvaluatingVolume(Vec3D scale, int res, float upperBound) {
    this(scale,res,res,res,upperBound);
  }
  
  public EvaluatingVolume(Vec3D scale, int RES, int resY, int resZ, float upperBound) {
    super(scale, RES, resY, resZ);
    this.upperBound = upperBound;
  }
  
   public void clear() {
    // muss implementiert sein
  }
  
  public final float getVoxelAt(int i) {
    X=   i/(RES*RES);//print(X+"   ");
    Y = (i-X*RES*RES)/RES; //print(Y+"   ");
    Z = (i-X*RES*RES)- Y*RES; 
    return getVoxelAt(X, Y, Z);
  }

  public final float getVoxelAt(int x, int y, int z) {
       float val; 
        //Je kleiner der Faktor (hier 20) desto größer erscheint der Kugelradius           
       float x1 =16*((float) x / (float) (RES-1)  - 0.5); //RES = 256,128,64,32,16
       float y1 =16*((float) y / (float) (RES-1)  - 0.5); 
       float z1 =16*((float) z / (float) (RES-1)  - 0.5);
       //print(x1,y1,z1+"  ");                 
       val = sq(sq(x1)+sq(y1)+sq(z1)-0.95*25)-0.8*(sq(z1-5)-2*sq(x1))*(sq(z1+5)-2*sq(y1));// print(val+"  ");   
       if (val > upperBound) {  
          val = 0;
        }
      return val;
  }
}
