
import toxi.geom.*;
import toxi.geom.mesh.*;
import toxi.math.*;
import toxi.volume.*;
import toxi.processing.*;

int RES = 128 ;//16,32,64,128,256
//Unterscheidung von innen - außen, also zwischen ISO und MAX_ISO:
float ISO =0.1;//0.000000000001;// ISO = 0 funktioniert mitunter nicht. 
float ISOStart = ISO;//Speichert den Startwert
float MAX_ISO = 2;// 2 UpperBound
boolean KZeichnen = false;//Koordinatenachsen nicht zeichnen

WETriangleMesh mesh;

ToxiclibsSupport gfx;
float currZoom = 0.5;
boolean isWireframe;
boolean inFarbe=false;
boolean lichtV,lichtH,lichtR,lichtL,lichtO,lichtU;


void setup() {
  size(800,800, P3D);
  colorMode(HSB);
  lichtH=true; lichtV=true; lichtO=true; lichtU=false; lichtR=false; lichtL=true;
  gfx = new ToxiclibsSupport(this);
  VolumetricSpace vol = new EvaluatingVolume(new Vec3D(800,800,800), RES, MAX_ISO);
  IsoSurface surface = new HashIsoSurface(vol);
  mesh = new WETriangleMesh();
  surface.computeSurfaceMesh(mesh, ISO);
 
}

void draw() {
  background(0);
  translate(width / 2, height / 2, 0);
  rotateX(mouseY * 0.01f);
  rotateY(mouseX * 0.01f);
  scale(currZoom);
  colorMode(RGB);//damit die Koordinatenachsen r,g,b gezeichnet werden
  if(KZeichnen) gfx.origin(500);//Koordinatenachsen
  colorMode(HSB);
  if (isWireframe) {
    noFill();
    stroke(255);
  } 
  else {
    fill(50,175,255); //gelb-grün
    noStroke();
    if(lichtV) directionalLight(0, 0, 255, 0, 0, -1); 
    if(lichtH) directionalLight(0, 0, 255, 0, 0,  1); 
    if(lichtR) directionalLight(0, 0, 255,-1, 0,  0); 
    if(lichtL) directionalLight(0, 0, 255, 1, 0,  0); 
    if(lichtO) directionalLight(0, 0, 255, 0,-1,  0);
    if(lichtU) directionalLight(0, 0, 255, 0, 1,  0); 
  } 
  if(!inFarbe) gfx.mesh(mesh, true);//hier wird gezeichnet //einfarbig   
   else gfx.meshNormalMapped(mesh, !isWireframe);//bunt
}

void voxelizeMesh() {
  if(ISO==0) ISO = 0.00001;//Funktioniert nicht mit ISO = 0
  mesh.computeFaceNormals();
  mesh.faceOutwards();
  mesh.computeVertexNormals();
  mesh.computeFaceNormals();
  mesh.faceOutwards();
  mesh.computeVertexNormals();
  MeshVoxelizer voxelizer=new MeshVoxelizer(RES);
  voxelizer.setWallThickness(0);//0,1,2
  VolumetricSpace vol = voxelizer.voxelizeMesh(mesh);
  vol.closeSides();
  IsoSurface surface = new HashIsoSurface(vol);
  mesh = new WETriangleMesh();
  surface.computeSurfaceMesh(mesh, ISO);
  mesh.computeVertexNormals();
  ISO = ISOStart;//zurücksetzen
}

void keyPressed() {
 if (key == 'w') {
    isWireframe = !isWireframe;
  }
  if (key == 'x') {
   voxelizeMesh();
  }  
  if (key == '-') {
    currZoom -= 0.1;
  }
  if (key == '+') {
    currZoom += 0.1;
  }
  if (key == 'l') {
    new LaplacianSmooth().filter(mesh, 1);
  }
  if (key == 'k') {
    KZeichnen= !KZeichnen;
  }
  if (key == 'f') {
    inFarbe= !inFarbe;
  }
  if (key == 's') {
     saveFrame("Bild-##.png");
  }
  if(key=='O') {lichtO=!lichtO;println("Licht oben: "+lichtO);}
  if(key=='U') {lichtU=!lichtU;println("Licht unten: "+lichtU);}
  if(key=='L') {lichtL=!lichtL;println("Licht links: "+lichtL);}
  if(key=='R') {lichtR=!lichtR;println("Licht rechts: "+lichtR);}
  if(key=='V') {lichtV=!lichtV;println("Licht vorn: "+lichtV);}
  if(key=='H') {lichtH=!lichtH;println("Licht hinten "+lichtH);}
  if (key == 'm') {
    beginShape(TRIANGLES);
      gfx.mesh( mesh );
    endShape();
    mesh.saveAsSTL( sketchPath( "cube.stl" ));
  }
}

class EvaluatingVolume extends VolumetricSpace {
  private int X,Y,Z;
  private float upperBound;
  
  public EvaluatingVolume(Vec3D scale, int res, float upperBound) {
    this(scale,res,res,res,upperBound);
  }
  
  public EvaluatingVolume(Vec3D scale, int RES, int resY, int resZ, float upperBound) {
    super(scale, RES, resY, resZ);
    this.upperBound = upperBound;
  }
  
   public void clear() {
    // muss implementiert sein
  }
  
  public final float getVoxelAt(int i) {
    X=   i/(RES*RES);//print(X+"   ");
    Y = (i-X*RES*RES)/RES; //print(Y+"   ");
    Z = (i-X*RES*RES)- Y*RES; 
    return getVoxelAt(X, Y, Z);
  }

  public final float getVoxelAt(int x, int y, int z) {
       float val; 
        //Je kleiner der Faktor (hier 1) desto größer        
       float u =2.3*((float) x / (float) (RES-1)  - 0.5); //RES = 256,128,64,32,16
       float v =2.3*((float) y / (float) (RES-1)  - 0.5); 
       float w =2.3*((float) z / (float) (RES-1)  - 0.5);//print(u+"  ");
                        
       val =0.1*abs( 2*v*(sq(v)-3*sq(u))*(1-sq(w))+sq((sq(u)+sq(v)))-(9*sq(w)-1)*(1-sq(w)) );
       if (val > upperBound) {  
          val = 0;
        }
      return val;
  }
}
