
import toxi.geom.*;
import toxi.geom.mesh.*;
import toxi.math.*;
import toxi.volume.*;
import toxi.processing.*;

int RES = 128 ;//16,32,64,128,256
//Unterscheidung von innen - außen, also zwischen ISO und MAX_ISO:
float ISO =0.0001;//0.000000000001;// ISO = 0 funktioniert mitunter nicht. 
float ISOStart = ISO;//Speichert den Startwert
float MAX_ISO = 10000;// 2 UpperBound - grad 14 - daher sehr schnell sehr groß!!
boolean KZeichnen = false;//Koordinatenachsen nicht zeichnen
boolean umZAchseDrehen = false;

WETriangleMesh mesh;

ToxiclibsSupport gfx;
float currZoom = 0.8;
boolean isWireframe;
boolean inFarbe=false;
boolean lichtV,lichtH,lichtR,lichtL,lichtO,lichtU;
float a,b;

void setup() {
  a =  -0.3;
  b =  -0.2;
  size(800,800, P3D);
  colorMode(HSB);
  lichtH=true; lichtV=true; lichtO=true; lichtU=false; lichtR=false; lichtL=true;
  gfx = new ToxiclibsSupport(this);
  VolumetricSpace vol = new EvaluatingVolume(new Vec3D(800,800,800), RES, MAX_ISO);
  IsoSurface surface = new HashIsoSurface(vol);
  mesh = new WETriangleMesh();
  surface.computeSurfaceMesh(mesh, ISO);
 
}

void draw() {
  background(0);
  translate(width / 2, height / 2, 0);
  rotateX(mouseY * 0.01f);
  rotateY(mouseX * 0.01f);
  if(umZAchseDrehen){
    rotateZ(1.5);
  }
  scale(currZoom);
  colorMode(RGB);//damit die Koordinatenachsen r,g,b gezeichnet werden
  if(KZeichnen) gfx.origin(500);//Koordinatenachsen
  colorMode(HSB);
  if (isWireframe) {
    noFill();
    stroke(255);
  } 
  else {
    fill(240,220,255); //rot (+blau)
    noStroke();
   if(lichtV) directionalLight(0, 0, 255, 0, 0, -1); 
   if(lichtH) directionalLight(0, 0, 255, 0, 0,  1); 
   if(lichtR) directionalLight(0, 0, 255,-1, 0,  0); 
   if(lichtL) directionalLight(0, 0, 255, 1, 0,  0); 
   if(lichtO) directionalLight(0, 0, 255, 0,-1,  0);
   if(lichtU) directionalLight(0, 0, 255, 0, 1,  0); 
  }
  if(!inFarbe) gfx.mesh(mesh, true);//hier wird gezeichnet //einfarbig   
   else gfx.meshNormalMapped(mesh, !isWireframe);//bunt
}

void voxelizeMesh() {
  if(ISO==0) ISO = 0.00001;//Funktioniert nicht mit ISO = 0
  mesh.computeFaceNormals();
  mesh.faceOutwards();
  mesh.computeVertexNormals();
  mesh.computeFaceNormals();
  mesh.faceOutwards();
  mesh.computeVertexNormals();
  MeshVoxelizer voxelizer=new MeshVoxelizer(RES);
  voxelizer.setWallThickness(0);//0,1,2
  VolumetricSpace vol = voxelizer.voxelizeMesh(mesh);
  vol.closeSides();
  IsoSurface surface = new HashIsoSurface(vol);
  mesh = new WETriangleMesh();
  surface.computeSurfaceMesh(mesh, ISO);
  mesh.computeVertexNormals();
  ISO = ISOStart;//zurücksetzen
}

void keyPressed() {
  if (key == 'w') {
    isWireframe = !isWireframe;
  }
  if (key == 'x') {
   voxelizeMesh();
  }  
  if (key == '-') {
    currZoom -= 0.1;
  }
  if (key == '+') {
    currZoom += 0.1;
  }
  if (key == 'l') {
    new LaplacianSmooth().filter(mesh, 1);
  }
  if (key == 'k') {
    KZeichnen= !KZeichnen;
  }
  if (key == 'f') {
    inFarbe= !inFarbe;
  }
  if (key == 'r') {
    umZAchseDrehen = true;
  }
  if (key == 's') {
     saveFrame("Bild-##.png");
  }
  if(key=='O') {lichtO=!lichtO;println("Licht oben: "+lichtO);}
  if(key=='U') {lichtU=!lichtU;println("Licht unten: "+lichtU);}
  if(key=='L') {lichtL=!lichtL;println("Licht links: "+lichtL);}
  if(key=='R') {lichtR=!lichtR;println("Licht rechts: "+lichtR);}
  if(key=='V') {lichtV=!lichtV;println("Licht vorn: "+lichtV);}
  if(key=='H') {lichtH=!lichtH;println("Licht hinten "+lichtH);}
  if (key == 'm') {
    beginShape(TRIANGLES);
      gfx.mesh( mesh );
    endShape();
    mesh.saveAsSTL( sketchPath( "cube.stl" ));
  }
}

class EvaluatingVolume extends VolumetricSpace {
  private int X,Y,Z;
  private float upperBound;
  
  public EvaluatingVolume(Vec3D scale, int res, float upperBound) {
    this(scale,res,res,res,upperBound);
  }
  
  public EvaluatingVolume(Vec3D scale, int RES, int resY, int resZ, float upperBound) {
    super(scale, RES, resY, resZ);
    this.upperBound = upperBound;
  }
  
   public void clear() {
    // muss implementiert sein
  }
  
  public final float getVoxelAt(int i) {
    X=   i/(RES*RES);//print(X+"   ");
    Y = (i-X*RES*RES)/RES; //print(Y+"   ");
    Z = (i-X*RES*RES)- Y*RES; 
    return getVoxelAt(X, Y, Z);
  }

  public final float getVoxelAt(int x, int y, int z) {
       float val; 
        //Je kleiner der Faktor (hier 1) desto größer        
       float u =3*((float) x / (float) (RES-1)  - 0.5); //RES = 256,128,64,32,16
       float v =3*((float) y / (float) (RES-1)  - 0.5); 
       float w =3*((float) z / (float) (RES-1)  - 0.5);//print(u+"  ");
                        
       val = sq(-8*sq(u*u + v*v)*(u*u + v*v + 1 + w*w + a*a - b*b)+4*a*a*(2*sq(u*u + v*v)- 
            (u*u*u - 3*u*v*v)*(u*u + v*v + 1))+
            8*a*a*(3*u*u*v - v*v*v)*w + 4*a*a*(u*u*u - 3*u*v*v)*w*w)-
            (u*u+v*v)*sq((2*(u*u+v*v)*sq(u*u+v*v+1+w*w+a*a-b*b)+8*sq(u*u+v*v)+
            4*a*a*(2*(u*u*u - 3*u*v*v) - (u*u + v*v)*(u*u + v*v + 1))-
            8*a*a*(3*u*u*v - v*v*v)*w - 4*(u*u + v*v)*a*a*w*w)); //Trefoil implicit
       if (val > upperBound) {  
          val = 0;
        }
      return val;
  }
}
